#include <iomanip>
#include <stdio.h>
#include <string.h>
#include "Grid.h"
#include "Grid_configure.cpp"
#include "Grid_load_power_spectra.cpp"
#include "Grid_generate_random_fields.cpp"
#include "Grid_output.cpp"
//#include "/home/moon/oliverf/Desktop/Shear_Programme/cfitsio/fitsio.h"

using namespace std;


/* _____ Content Grid.cpp
 * 
 * ..... 1. Constructor
 * ..... 2. Destructor
 * ..... 3. check_l_range
 * ..... 4. set_number_of_threads
 * 
 */


/*******************************************************************************************************************************************************
 * 1. Constructor
 * 
 * Description:
 * - Calls functions to read the configuration files.
 * - Checks whether the given l_max value is small enough, such that
 *   from the Fourier transformed grid a sub grid of edge length
 *   theta_max can be cut out.
 * - Calls the functions that initialize and build the random fields.
 * 
 * Arguments:
 * 
 * - seed: seed values of the RNG when drawing the Fourier modes
 * 
*******************************************************************************************************************************************************/

Grid::Grid(int seed){
  
  cout << "\nStarting logsim:\n\n";

  
  // Tolerance region around the spatial grid, in which no galaxies will be placed,
  // but which will be used for interpolation:
  this->tolerance = 8;
  
  this->configure("configure_general");
  this->configure_delta_fields("configure_clustered_galaxies");
  this->configure_shear_fields("configure_shear_fields");
  this->read_power_spectra("power_spectra");
  this->log_normal_tranform_to_power_spectra();

  this->seed_value = seed;
  
  this->check_l_range();
  this->set_number_of_threads();
  this->show_status();
  this->set_grid();
  cout.flush();
  
  
}

/*******************************************************************************************************************************************************
 * 2. Destructor
 * Description:
 * -
 * Arguments:
 * -
 * 
*******************************************************************************************************************************************************/

Grid::~Grid(){
}

/*******************************************************************************************************************************************************
 * 3. check_l_range
 * Description:
 * - Checks whether specified l_max value is compatible with demanded theta_max.
 * - Prints the resolution on Fourier space.
 * 
 * Arguments:
 * -
 * 
*******************************************************************************************************************************************************/

void Grid::check_l_range(){
  
  data_type dl = 2*this->l_max/data_type(this->number_of_steps);
  this->l_min = dl;
  
  cout << "resolution in l-space: " << this->l_min << '\n';
  
  // Update l_max if Fourier transformed grid would be smaller than 
  // demanded by theta_max:
  if(this->l_max > constants::pi*data_type(this->number_of_steps)/(this->theta_max*constants::arcmin)){
    this->l_max = constants::pi*data_type(this->number_of_steps)/(this->theta_max*constants::arcmin);
    cout << "l_max value too high, updated to " << int(this->l_max) << '\n';
  }
  else
    cout << "l_max value approved\n"; 
}

/*******************************************************************************************************************************************************
 * 4. set_number_of_threads
 * Description:
 * - Checks whether specified number of threads is a power of two.
 *   If not, it reduces number_of_threads to the next lowest power of 2
 *
 * Arguments:
 * -
 * 
*******************************************************************************************************************************************************/

void Grid::set_number_of_threads(){
  
  int n = 1;
  while(n*2 <= this->number_of_threads)
    n*=2;
  
  if(n != this->number_of_threads){
    cout << "Number of threads updated to " << n << "\n\n";
    this->number_of_threads = n;
  }
  
  
}
