#include "Hankel_Transformer.cpp"

/* _____ Content Grid_load_power_spectra.cpp
 * 
 * ..... 1. read_power_spectra
 * ..... 2. log_normal_tranform_to_power_spectra
 * 
 */

/*******************************************************************************************************************************************************
 * 1. read_power_spectra
 * Description:
 * - Reads power spectra. For two fields the format of the table should e.g. be:
 *   P11 P12 P22
 *   For three fields it should be:
 *   P11 P12 P13 P22 P23 P33
 *   etc.
 * Arguments:
 * - power_spectrum_name: name of file containing the input power spectra
 * 
*******************************************************************************************************************************************************/


void Grid::read_power_spectra(string power_spectrum_name){

  int n_delta = this->number_of_delta_fields;
  int n_shear = this->number_of_shear_fields;
  int n_fields = n_shear + n_delta;
  int n_ell;

  double current_ell;

  vector<data_type> ell_values_input(0, 0.0);
  vector<vector<data_type> > current_line(n_fields, vector<data_type>(n_fields, 0.0));
  vector<vector<vector<data_type> > > power_spectra_aux(n_fields, vector<vector<data_type> >(n_fields, vector<data_type>(0, 0.0)));

  fstream input;
  input.open(power_spectrum_name.c_str());
  string first_line;
  getline(input, first_line);
  
  while(input.good()){
    
    input >> current_ell;
    
    for(int i = 0; i < n_fields; i++){
      for(int j = i; j < n_fields; j++){
	input >> current_line[i][j];
      }
    }
    
    if(input.good()){
      ell_values_input.push_back(log(current_ell));
      for(int i = 0; i < n_fields; i++){
	for(int j = i; j < n_fields; j++){
	  power_spectra_aux[i][j].push_back(current_line[i][j]);
	}
      }
    }
  }
  n_ell = ell_values_input.size();
  
  
  this->ln_l = valarray<data_type>(0.0, 1024);
  this->power_spectra = vector<vector<valarray<data_type> > >(n_fields, vector<valarray<data_type> >(n_fields, valarray<data_type>(0.0, 1024)));
  double dln_l = pow(2.0e6, 1.0/1023.0);
  for(int i = 0; i < 1024; i++){
    data_type ell = 0.5*data_type(pow(dln_l, double(i)));
    data_type log_ell = log(ell);
    this->ln_l[i] = log_ell;
  }

  valarray<data_type> interpolation_ell(0.0, n_ell);
  valarray<data_type> interpolation_P(0.0, n_ell);

  for(int l = 0; l < n_ell; l++)
    interpolation_ell[l] = ell_values_input[l];

  for(int i = 0; i < n_fields; i++){
    for(int j = i; j < n_fields; j++){
      for(int l = 0; l < n_ell; l++)
	interpolation_P[l] = power_spectra_aux[i][j][l];
      
      for(int l = 0; l < 1024; l++){
	data_type ell = 0.5*data_type(pow(dln_l, double(i)));
	data_type pow;
	if(this->ln_l[l] < interpolation_ell[0] || this->ln_l[l] > interpolation_ell[n_ell-1])
	  pow = 0.0;
	else
	  pow = interpolate_neville_aitken(this->ln_l[l], &interpolation_ell, &interpolation_P, 2);
       if(pow < 0.0)
         pow = 0.0;
	this->power_spectra[i][j][l] = pow;
	this->power_spectra[j][i][l] = pow;
      }
    }
  }
  
} 


/*******************************************************************************************************************************************************
 * 2. log_normal_tranform_to_power_spectra
 * Description:
 * - Tranforms the power spectra involving log-normal fields to the power spectra
 *   that involve the corresponding Gaussian fields.
 * Arguments:
 * - 
 * 
*******************************************************************************************************************************************************/


void Grid::log_normal_tranform_to_power_spectra(){
  
  int n_delta = this->number_of_delta_fields;
  int n_shear = this->number_of_shear_fields;
  int n_fields = n_shear + n_delta;
  int n_ell = this->ln_l.size();
  int log_normal_transform;
  int k,m;
  
  data_type minimal_value;
  
  vector<int> log_normal_field(n_fields, 0);
  valarray<double> aux_ln_l(0.0, n_ell);
  valarray<double> aux_power_spectrum(0.0, n_ell);
  vector<data_type> minimal_values(n_fields, 0.0);
  
  for(int i = 0; i < n_delta; i++){
    log_normal_field[i] = this->log_normal_delta[i];
    if(log_normal_field[i])
      minimal_values[i] = 1.0;
    else
      minimal_values[i] = 0.0;
  }
  for(int i = n_delta; i < n_fields; i++){
    log_normal_field[i] = this->log_normal_shear[i-n_delta];
    if(log_normal_field[i])
      minimal_values[i] = this->kappa_zeros[i-n_delta];
    else
      minimal_values[i] = 0.0;
  }
  
  
  for(int i = 0; i < n_fields; i++){
    for(int j = i; j < n_fields; j++){
      log_normal_transform = log_normal_field[i]*log_normal_field[j];
      minimal_value = minimal_values[i]*minimal_values[j];
      minimal_value = sqrt(minimal_value);
      
      if(log_normal_transform == 1){
	for(int l = 0; l < n_ell; l++){
	  aux_power_spectrum[l] = double(this->power_spectra[i][j][l]);
	  aux_ln_l[l] = double(this->ln_l[l]);
	}
	Hankel_Transformer transformer(1, 0, aux_ln_l, aux_power_spectrum);
	transformer.lognorm_to_Gaussian_P(minimal_value, &aux_power_spectrum);
	for(int l = 0; l < n_ell; l++){
	  this->power_spectra[i][j][l] = data_type(aux_power_spectrum[l]);
	  this->power_spectra[j][i][l] = data_type(aux_power_spectrum[l]);
	}
      }
      
      if(log_normal_field[i] == 0 && log_normal_field[j] == 1){
	for(int l = 0; l < n_ell; l++)
	  this->power_spectra[i][j][l] /= minimal_values[j];
	this->power_spectra[j][i] = this->power_spectra[i][j];
      }
      
      if(log_normal_field[i] == 1 && log_normal_field[j] == 0){
	for(int l = 0; l < n_ell; l++)
	  this->power_spectra[i][j][l] /= minimal_values[i];
	this->power_spectra[j][i] = this->power_spectra[i][j];
      }
      
      for(int l = 0; l < n_ell; l++){
	if(this->power_spectra[i][j][l]<0.0){
	  this->power_spectra[i][j][l] = 0.0;
	  this->power_spectra[j][i][l] = 0.0;
	}
      }
    }
  }
} 
