#include <fitsio.h>

/*******************************************
 *******************************************
 **_____________ 2. OUTPUT _______________**
 *******************************************
 ******************************************* 
 *                                         *
 * ..... 1. print_shear_catalog            *
 * ..... 2. print_clustered_galaxies       *
 * ..... 3. print_random_points            *
 * ..... 4. print_shear_grid               *
 * ..... 5. print_delta_grid               *
 * ..... 6. show_status                    *
 *                                         *
 *******************************************
 *******************************************/


/*******************************************************************************************************************************************************
 * 1. print_shear_catalog
 * Description:
 * - Prints a shape catalog.
 * Arguments:
 * - n: number of galaxies
 * - sigma_eps: ellipticity dispersion per component
 * - n_of_grid: shear-field onto which the galaxies should be placed
 *   (counting starts from 1)
 * - output_name: name of the output file.
 * 
*******************************************************************************************************************************************************/

void Grid::print_shear_catalog(int n, data_type sigma_eps, int n_of_grid, string output_name){
  cout << "Generating shape catalog on shear field " << n_of_grid << ":\n";
  cout << fixed << setprecision(2);
  cout << "Ellipticity dispersion per component: " << sigma_eps << '\n';
  
  bool intrinsic_noise = 0;
  if(sigma_eps >= constants::eps)
    intrinsic_noise = 1;
  
  int index;
  int output;
  int status = 0;
  int ncol = 4;
  int row = 0;
  int number_of_grid = n_of_grid-1+this->number_of_delta_fields;
  
  string file_name = "!"+output_name;
  
  float ra;
  float de;
  float s1;
  float s2;
  data_type l = constants::arcmin*data_type(this->number_of_steps)/(2*this->l_max);
  mt19937_64 gen(2);
  data_type normalization = gen.max();
    
  char *ttype[ncol];
  char *tform[ncol];
  char *tunit[ncol];
  string dummy;
  cout.flush();
  ttype[0] = "ra"; tform[0] = "1E"; tunit[0] = "arcmin";
  ttype[1] = "dec"; tform[1] = "1E"; tunit[1] = "arcmin";
  ttype[2] = "s1"; tform[2] = "1E"; tunit[2] = "/";
  ttype[3] = "s2"; tform[3] = "1E"; tunit[3] = "/";
  
  fitsfile *fptr;
  fits_create_file(&fptr, file_name.c_str(), &status);
  fits_create_tbl(fptr, BINARY_TBL, n, ncol, ttype, tform, tunit, "mock_shear_catalog", &status);
  fits_report_error(stderr, status);
  
  vector<thread> th;
  
  this->set_sub_grids(n, number_of_grid, intrinsic_noise);
  cout.flush();
  
  for(int i = 0; i < this->parts_ra; i++){
    for(int j = 0; j < this->parts_de; j++){
      index = i + j*this->parts_ra;
      output = index-this->number_of_threads/2;
      
      
      
      th.push_back(thread(set_galaxy_data, &this->sub_positions[index], &this->sub_random_numbers[index],
			 &this->sub_thetas_ra[i], &this->sub_thetas_de[j], 
			 &this->sub_grids[index], &this->sub_galaxy_data[index],
			 intrinsic_noise, sigma_eps, output, normalization));
    }
  }

    
  for(int i = 0; i < this->number_of_threads; i++)
    th[i].join();
  
  for(int i = 0; i < this->number_of_threads; i++){
    cout << "\rWriting catalog:                " << 100.0*data_type(i)/data_type(this->number_of_threads) << "%";
    cout.flush();
    
    for(int j = 0; j < this->sub_galaxy_data[i].size(); j++){
      row++;
      
      //switched ra-dec as a test!!
      ra = float(this->sub_positions[i][2*j]);
      de = float(this->sub_positions[i][2*j+1]-this->theta_max/2);
      s1 = float(this->sub_galaxy_data[i][j][0]);
      s2 = float(this->sub_galaxy_data[i][j][1]);
      
      fits_write_col(fptr, TFLOAT, 1, row, 1, 1, &ra, &status);
      fits_write_col(fptr, TFLOAT, 2, row, 1, 1, &de, &status);
      fits_write_col(fptr, TFLOAT, 3, row, 1, 1, &s1, &status);
      fits_write_col(fptr, TFLOAT, 4, row, 1, 1, &s2, &status);
    }
  }
  
  fits_report_error(stderr, status);
  cout << "\rWriting catalog:                100%     \n";
  fits_close_file(fptr, &status);
  fits_report_error(stderr, status);

  this->sub_positions.clear();
  this->sub_grids.clear();
  this->sub_galaxy_data.clear();
  this->sub_thetas_ra.clear();
  this->sub_thetas_de.clear();

}

/*******************************************************************************************************************************************************
 * 2. print_clustered_galaxies
 * Description:
 * - Prints a catalog of clustered galaxies.
 * Arguments:
 * - n: number of galaxies
 * - bias: coefficient of the linear bias term in delta_N = b*delta + c*delta^2
 * - bias_quadratic: coefficient of the quadratic term in delta_N = b*delta + c*delta^2
 * - n_of_grid: delta-field onto which the galaxies should be placed
 *   (counting starts from 1)
 * - output_name: name of the output file.
 * 
*******************************************************************************************************************************************************/


void Grid::print_clustered_galaxies(int n, data_type bias, data_type bias_quadratic, int n_of_grid, string output_name){

  cout << "Generating catalog of clustered galaxies on delta field " << n_of_grid << ":\n";
  cout << "Linear bias: " << bias << '\n';
  cout << "Quadratic bias: " << bias_quadratic << '\n';

  if(bias_quadratic > pow(bias, 2.0)){
    cerr << "Quadratic bias term for field is too large.\n";
    exit(EXIT_FAILURE);
  }
  if(bias_quadratic != 0.0 && this->log_normal_delta[n_of_grid - 1] == 1){
    cerr << "Warning: quadratic bias in log-normal mode is still experimental!\n";
  }
  
  int status = 0;
  int n_new = 0;
  int row = 0;
  int seed = this->seed_value;
  this->seed_value = next_prime(this->seed_value);
  this->seed_value = next_prime(this->seed_value);
  int number_of_grid = n_of_grid - 1;
  int n_grid = this->random_fields[number_of_grid].size();
  int n_sub = n; n_sub *= 5; n_sub /= n_grid;
  if(n_sub < 3000) n_sub = 3000;
  
  float ra, ra_trough;
  float de, de_trough;
  float s1 = 0.0;
  float s2 = 0.0;
  
  string file_name = "!"+output_name;
  
  data_type dtheta = constants::pi/this->l_max;
  dtheta /= constants::arcmin;
  data_type n_bar = data_type(n)/pow(data_type(n_grid), data_type(2.0));
  
  mt19937_64 gen(seed);
  valarray<valarray<int> > galaxy_numbers(valarray<int>(0, n_grid), n_grid);
  valarray<valarray<data_type> > random_numbers(valarray<data_type>(0.0, n_sub), n_grid);
  for(int i = 0; i < n_grid; i++){
    for(int j = 0; j < n_sub; j++){
      random_numbers[i][j] = data_type(gen())/data_type(gen.max());
    }
  }
  
  // initializing the positions of the galaxies and
  // distributing them to the ra and de sub grids:
  
  vector<vector<data_type> > fore_ground_positions = vector<vector<data_type> >(this->number_of_threads, vector<data_type>(0, 0.0));
  
  for(int i = 0; i < n_grid; i+= this->number_of_threads){
    if(i%1000 == 0){
      cout << "\rPlacing clustered galaxies:   " << 100.0*data_type(i)/data_type(n) << "%";
      cout.flush();
    }

    vector<thread> th;
    for(int j = 0; j < min(this->number_of_threads, n_grid-i); j++)
      th.push_back(thread(place_clustered_galaxies, &fore_ground_positions[j], &this->random_fields[number_of_grid][i+j], &galaxy_numbers[i+j], dtheta, n_bar, bias, bias_quadratic, this->variances_of_Gaussian_delta_fields[number_of_grid], i+j, this->tolerance, this->log_normal_delta[number_of_grid], &random_numbers[i+j]));
    for(int j = 0; j < min(this->number_of_threads, n_grid-i); j++)
      th[j].join();
  }
  
  for(int j = 0; j < this->number_of_threads; j++)
    n_new += fore_ground_positions[j].size()/2;
  
  char *ttype[4];
  char *tform[4];
  char *tunit[4];
  string dummy;
  cout.flush();
  ttype[0] = "ra"; tform[0] = "1E"; tunit[0] = "arcmin";
  ttype[1] = "dec"; tform[1] = "1E"; tunit[1] = "arcmin";
  ttype[2] = "s1"; tform[2] = "1E"; tunit[2] = "/";
  ttype[3] = "s2"; tform[3] = "1E"; tunit[3] = "/";
  
  fitsfile *fptr;
  fits_create_file(&fptr, file_name.c_str(), &status);
  fits_create_tbl(fptr, BINARY_TBL, n_new, 4, ttype, tform, tunit, "mock_foreground_catalog", &status);
  fits_report_error(stderr, status);
  
  vector<thread> th;
  
  cout << "\rPlacing clustered galaxies:   100%     \n";
  cout.flush();
  
  for(int i = 0; i < this->number_of_threads; i++){
    cout << "\rWriting catalog:                " << 100.0*data_type(i)/data_type(this->number_of_threads) << "%";
    cout.flush();
    
    for(int j = 0; j < fore_ground_positions[i].size()/2; j++){
      row++;
      
      ra = float(fore_ground_positions[i][2*j]);
      de = float(fore_ground_positions[i][2*j+1]-this->theta_max/2);
      
      fits_write_col(fptr, TFLOAT, 1, row, 1, 1, &ra, &status);
      fits_write_col(fptr, TFLOAT, 2, row, 1, 1, &de, &status);
      fits_write_col(fptr, TFLOAT, 3, row, 1, 1, &s1, &status);
      fits_write_col(fptr, TFLOAT, 4, row, 1, 1, &s2, &status);
    }
  }

  fits_report_error(stderr, status);
  cout << "\rWriting catalog:                 100%       \n\n";
  fits_close_file(fptr, &status);
  fits_report_error(stderr, status);  
  /*
  data_type trough_radius = 10.0;
  int integer_radius = int(trough_radius/dtheta +0.5);
  int integer_radius_sq = integer_radius*integer_radius;
  int start_index = this->tolerance + integer_radius + 1;
  int max_index = n_grid - this->tolerance - integer_radius - 1;
  int max_count = 200;
  double mean_counts = 0.0;
  int n_trough = 0;
  vector<data_type> trough_positions;
  
  for(int i = start_index; i < max_index; i+= 6){
      cout << "\rSearching troughs:            " << 100.0*data_type(i)/data_type(max_index-start_index) << "%";
    for(int j = start_index; j < max_index; j+=6){
    
      int counts = 0;
      
      for(int k = -integer_radius; k <= integer_radius; k++){
	for(int l = -integer_radius; l <= integer_radius; l++){
	  if(k*k+l*l <= integer_radius_sq){
	    counts += galaxy_numbers[i+k][j+l];
	  }
        }
      }
      
      if(counts < max_count){
        ra = data_type(i-this->tolerance)*dtheta;
        de = data_type(j-this->tolerance)*dtheta-this->theta_max/2.0;
	trough_positions.push_back(ra);
	trough_positions.push_back(de);
	mean_counts += double(counts);
	n_trough++;
      }
      
    }
  }
  n_new = trough_positions.size()/2;
  cout << "\n";
  cout << n_trough << "troughs found.\n";
  cout << "Mean counts in troughs: " << mean_counts/double(n_trough) << "\n";
  fitsfile *fptr2;
  fits_create_file(&fptr2, "!./troughs.fits", &status);
  fits_create_tbl(fptr2, BINARY_TBL, n_new, 4, ttype, tform, tunit, "mock_trough_catalog", &status);
  fits_report_error(stderr, status);
  
  
  
  row = 0;

  for(int i = 0; i < this->number_of_threads; i++){
    cout << "\rWriting trough catalog " << number_of_grid << ":       " << 100.0*data_type(i)/data_type(this->number_of_threads) << "%";
    cout.flush();
    
    for(int j = 0; j < n_new; j++){
      row++;
      
      ra = trough_positions[2*j];
      de = trough_positions[2*j+1];
      s1 = 0.0;
      s2 = 0.0;
      
      fits_write_col(fptr2, TFLOAT, 1, row, 1, 1, &ra, &status);
      fits_write_col(fptr2, TFLOAT, 2, row, 1, 1, &de, &status);
      fits_write_col(fptr2, TFLOAT, 3, row, 1, 1, &s1, &status);
      fits_write_col(fptr2, TFLOAT, 4, row, 1, 1, &s2, &status);
    }
  }

  fits_report_error(stderr, status);
  cout << "\rWriting trough catalog " << number_of_grid << ":       100.0%\n";
  fits_close_file(fptr2, &status);
  fits_report_error(stderr, status);*/
  
}


/*******************************************************************************************************************************************************
 * 3. print_random_points
 * Description:
 * - Prints a catalog of random points.
 * Arguments:
 * - number_of_randoms: number of random points
 * - output_name: name of the output file.
 * 
*******************************************************************************************************************************************************/



void Grid::print_random_points(int number_of_randoms, string output_name){

  int status = 0;
  int seed = this->seed_value;
  
  float ra;
  float de;
  
  string file_name = "!"+output_name;
  
  mt19937_64 gen(seed);
  
  char *ttype[2];
  char *tform[2];
  char *tunit[2];
  string dummy;
  cout.flush();
  ttype[0] = "ra"; tform[0] = "1E"; tunit[0] = "arcmin";
  ttype[1] = "dec"; tform[1] = "1E"; tunit[1] = "arcmin";
  
  fitsfile *fptr;
  fits_create_file(&fptr, file_name.c_str(), &status);
  fits_create_tbl(fptr, BINARY_TBL, number_of_randoms, 2, ttype, tform, tunit, "mock_foreground_catalog", &status);
  fits_report_error(stderr, status);
    
  cout << "Placing foreground galaxies:  100%   \n";
  cout.flush();
  
  for(int row = 0; row < number_of_randoms; row++){
    if(row%1000 == 0){
      cout << "\rWriting random catalog:         " << 100.0*data_type(row)/data_type(number_of_randoms) << "%";
      cout.flush();
    }
    ra = float(this->theta_max*data_type(gen())/data_type(gen.max()));
    de = float(this->theta_max*data_type(gen())/data_type(gen.max()) - 0.5*this->theta_max);
      
    fits_write_col(fptr, TFLOAT, 1, row+1, 1, 1, &ra, &status);
    fits_write_col(fptr, TFLOAT, 2, row+1, 1, 1, &de, &status);
  }
  
  fits_report_error(stderr, status);
  cout << "\rWriting random catalog:         100%     \n\n";
  fits_close_file(fptr, &status);
  fits_report_error(stderr, status);
  
}



/*******************************************************************************************************************************************************
 * 4. print_shear_grid
 * Description:
 * -
 * Arguments:
 * -
 * 
*******************************************************************************************************************************************************/


void Grid::print_shear_grid(string output_name, int number_of_grid){
  
  int n = this->random_fields[number_of_grid+this->number_of_delta_fields].size();
  int nth = this->thetas.size();
  int index;
  int output;
  int status = 0;
  int ncol = 4;
  long int row = 0;
  
  float ra;
  float de;
  float s1;
  float s2;
  
  string file_name = "!"+output_name;
  
  data_type l = constants::arcmin*data_type(this->number_of_steps)/(2*this->l_max);
  data_type dtheta = constants::pi/(constants::arcmin*this->l_max);
    
  char *ttype[ncol];
  char *tform[ncol];
  char *tunit[ncol];
  string dummy;

  ttype[0] = "ra"; tform[0] = "1E"; tunit[0] = "arcmin";
  ttype[1] = "dec"; tform[1] = "1E"; tunit[1] = "arcmin";
  ttype[2] = "s1"; tform[2] = "1E"; tunit[2] = "/";
  ttype[3] = "s2"; tform[3] = "1E"; tunit[3] = "/";
  
  fitsfile *fptr;
  fits_report_error(stderr, status);
  fits_create_file(&fptr, file_name.c_str(), &status);
  fits_report_error(stderr, status);
  fits_create_tbl(fptr, BINARY_TBL, n, ncol, ttype, tform, tunit, "mock_shear_catalog", &status);
  fits_report_error(stderr, status);

  
  for(int i = 0; i < n; i++){
    cout << "\rPrinting shear grid " << number_of_grid << ":          " << 100.0*data_type(i)/n << "%  ";
    cout.flush();
    for(int j = 0; j < n; j++){
      row++;
	
      ra = float(this->thetas[i]);
      de = float(this->thetas[j]-this->thetas[nth-1]/2.0);
      s1 = float(this->random_fields[number_of_grid+this->number_of_delta_fields][i][j].real());
      s2 = float(this->random_fields[number_of_grid+this->number_of_delta_fields][i][j].imag());
	  
      fits_write_col(fptr, TFLOAT, 1, row, 1, 1, &ra, &status);
      fits_write_col(fptr, TFLOAT, 2, row, 1, 1, &de, &status);
      fits_write_col(fptr, TFLOAT, 3, row, 1, 1, &s1, &status);
      fits_write_col(fptr, TFLOAT, 4, row, 1, 1, &s2, &status);
    }
  }
  
  fits_report_error(stderr, status);
  cout << "\rPrinting shear grid " << number_of_grid << ":          100%     \n\n";
  fits_close_file(fptr, &status);
  fits_report_error(stderr, status);
}


/*******************************************************************************************************************************************************
 * 5. print_delta_grid
 * Description:
 * -
 * Arguments:
 * -
 * 
*******************************************************************************************************************************************************/


void Grid::print_delta_grid(string output_name, int number_of_grid){
  
  int n = this->random_fields[number_of_grid].size();
  int nth = this->thetas.size();
  int index;
  int output;
  int status = 0;
  int ncol = 3;
  long int row = 0;
  
  float ra;
  float de;
  float delta;
  
  string file_name = "!"+output_name;
  
  data_type l = constants::arcmin*data_type(this->number_of_steps)/(2*this->l_max);
  data_type dtheta = constants::pi/(constants::arcmin*this->l_max);
  
  complex<data_type> shear(0,0);
  
  char *ttype[ncol];
  char *tform[ncol];
  char *tunit[ncol];
  string dummy;

  ttype[0] = "ra"; tform[0] = "1E"; tunit[0] = "arcmin";
  ttype[1] = "dec"; tform[1] = "1E"; tunit[1] = "arcmin";
  ttype[2] = "delta"; tform[2] = "1E"; tunit[2] = "/";
  
  fitsfile *fptr;
  fits_report_error(stderr, status);
  fits_create_file(&fptr, file_name.c_str(), &status);
  fits_report_error(stderr, status);
  fits_create_tbl(fptr, BINARY_TBL, n, ncol, ttype, tform, tunit, "mock_shear_catalog", &status);
  fits_report_error(stderr, status);

  
  for(int i = 0; i < n; i++){
    cout << "\rPrinting delta grid " << number_of_grid << ":          " << 100.0*data_type(i)/n << "%  ";
    cout.flush();
    for(int j = 0; j < n; j++){
      row++;
	
      ra = float(this->thetas[i]);
      de = float(this->thetas[j]-this->thetas[nth-1]/2.0);
      delta = float(this->random_fields[number_of_grid][i][j].real());
	  
      fits_write_col(fptr, TFLOAT, 1, row, 1, 1, &ra, &status);
      fits_write_col(fptr, TFLOAT, 2, row, 1, 1, &de, &status);
      fits_write_col(fptr, TFLOAT, 3, row, 1, 1, &delta, &status);
    }
  }
  
  fits_report_error(stderr, status);
  cout << "\rPrinting delta grid " << number_of_grid << ":          100%     \n\n";
  fits_close_file(fptr, &status);
  fits_report_error(stderr, status);
}


/*******************************************************************************************************************************************************
 * 6. show_status
 * Description:
 * -
 * Arguments:
 * -
 * 
*******************************************************************************************************************************************************/


void Grid::show_status(){
  
  cout << fixed << setprecision(2);
  cout << "\nSimulating random fields with:\n\n";
  cout << "               A = (" << this->theta_max << " arcmin)^2\n";
  cout << "           l_max = " << this->l_max << '\n';
  cout << "           l_min = " << this->l_min << '\n';
  cout << "          N_grid = (" << this->number_of_steps << ")^2\n";
  cout << "         N_delta = " << this->number_of_delta_fields << '\n';
  cout << "         N_shear = " << this->number_of_shear_fields << '\n';
  cout << "            seed = " << this->seed_value << '\n';
  cout << '\n';
  
}