using namespace std;

/*******************************************************************************************************************************************************
 * fft
 * 
 * Description:
 * - Fast Fourier transform of the array (*f).
 * 
 * Arguments:
 * - f: pointer on the array to be transformed
 * - e: stored exponential factor appearing in the Fourier transform.
 * 
 * Comments:
 *   - The exponential factor "e" is provided from outside such that
 *     the exponential function does not have to be computed in every
 *     call of fft.
 * 
*******************************************************************************************************************************************************/

void fft(valarray<complex<data_type> > *f, complex<data_type> e){
 
 
  int n2 = (*f).size();
  
  if(n2 > 1){
    
    int n = n2/2;
    
    complex<data_type> one = 1.0;
    complex<data_type> two = 1.0;
    complex<data_type> three = 1.0;
    
    valarray<complex<data_type> > f1(n);
    valarray<complex<data_type> > f2(n);
    
    f1 = (*f)[slice(0,n,2)];
    f2 = (*f)[slice(1,n,2)];
    
    fft(&f1, e*e);
    fft(&f2, e*e);
    
    for(int i = 0; i < n; i++){
      two = f1[i];
      three = one*f2[i];
      (*f)[i] =  two + three; 
      (*f)[i+n] = two - three;
      one *= e;
    }
    
    
  }
  
}




void fft_inverse(valarray<complex<data_type> > *f, complex<data_type> e, complex<data_type> normalization){
 
 
  int n2 = (*f).size();
  (*f) *= normalization;
  
  if(n2 > 1){
    
    int n = n2/2;
    
    complex<data_type> one = 1.0;
    complex<data_type> two = 1.0;
    complex<data_type> three = 1.0;
    
    valarray<complex<data_type> > f1(n);
    valarray<complex<data_type> > f2(n);
    
    f1 = (*f)[slice(0,n,2)];
    f2 = (*f)[slice(1,n,2)];
    
    fft(&f1, e*e);
    fft(&f2, e*e);
    
    for(int i = 0; i < n; i++){
      two = f1[i];
      three = one*f2[i];
      (*f)[i] =  two + three; 
      (*f)[i+n] = two - three;
      one *= e;
    }
    
    
  }
  
}

/*******************************************************************************************************************************************************
 * fft_aux
 * 
 * Description:
 * - auxilliary on order to make better cpu_usage
 * 
*******************************************************************************************************************************************************/

void fft_aux_1(int n, int start_index, valarray<valarray<complex<data_type> > > *f, complex<data_type> e){
  
  for(int i = start_index; i < start_index + n; i++){
    fft(&(*f)[i], e*e);
  }
  
}


void fft_aux_2(int n, int start_index, int n1, valarray<valarray<complex<data_type> > > *f, complex<data_type> e){
  
  valarray<complex<data_type> > g(complex<data_type>(0.0), n1);
  
  for(int i = start_index; i < start_index + n; i++){
    for(int k = 0; k < n1; k++)
	  g[k] = (*f)[k][i];
    fft(&g, e*e);
    for(int k = 0; k < n1; k++)
	  (*f)[k][i] = g[k];
  }
  
}


void fft_aux_1_inverse(int n, int start_index, valarray<valarray<complex<data_type> > > *f, complex<data_type> e, complex<data_type> normalization){
  
  for(int i = start_index; i < start_index + n; i++){
    fft_inverse(&(*f)[i], e*e, normalization);
  }
  
}



void fft_aux_2_inverse(int n, int start_index, int n1, valarray<valarray<complex<data_type> > > *f, complex<data_type> e, complex<data_type> normalization){
  
  valarray<complex<data_type> > g(complex<data_type>(0.0), n1);
  
  for(int i = start_index; i < start_index + n; i++){
    for(int k = 0; k < n1; k++)
	  g[k] = (*f)[k][i];
    fft_inverse(&g, e*e, normalization);
    for(int k = 0; k < n1; k++)
	  (*f)[k][i] = g[k];
  }
  
}

/*******************************************************************************************************************************************************
 * fft_2D
 * 
 * Description:
 * - two dimensional Fourier transform
 * 
 * Arguments:
 * - f: pointer of the array to be transformed
 * - n1 & n2: edge sizes of the two dimensional array in (*f)
 * - thread_number: number of thread into which the computation
 *                  shall be split
 * - output: variable governing the output of the method
 * 
 * Comments:
 *   - The number of threads should be a power of 2 or at least n1 and n2 should be divisible by it.
 *     Otherwise an exception is drawn.
 * 
*******************************************************************************************************************************************************/


void fft_2D(valarray<valarray<complex<data_type> > > *f, int n1, int n2, int thread_number, int output, int number_of_grid){
  
  int n_of_threads;
  int stack = 8;
  
  complex<data_type> im(0,constants::pi/data_type(n2));
  complex<data_type> e = exp(im);
  
  // Make sure, that n1 is divisible by number of threads.
  n_of_threads = thread_number;
  while(n1%n_of_threads != 0)
    n_of_threads++;
  
  {
    //thread th[n_of_threads];
    for(int i = 0; i < n1; i+=n_of_threads*stack){
      
      vector<thread> th;
      
      if(output == 1)
	cout << "\rGenerating delta field " << number_of_grid+1 << ":       "  << data_type(400*i/n1)/10 << "%";
      else if(output == 2)
	cout << "\rGenerating kappa field " << number_of_grid+1 << ":       "  << data_type(200*i/n1)/10 << "%";
      else if(output == 3)
	cout << "\rGenerating delta field " << number_of_grid+1 << ":       "  << data_type(500*i/n1)/10 << "%";
      else if(output == 4)
	cout << "\rGenerating shear field " << number_of_grid+1 << ":       "  << data_type(300*i/n1)/10 + 40 << "%";
      cout.flush();

    
      for(int j = 0; j < n_of_threads; j++){
        th.push_back(thread(fft_aux_1,stack, i+stack*j, f, e));
      }
      
    
      for(int j = 0; j < n_of_threads; j++){
        th[j].join();
      }
    
    }
  }
  
  im = complex<data_type>(0,constants::pi/data_type(n1));
  e = exp(im);
  // Make sure, that n1 is divisible by number of threads.
  n_of_threads = thread_number;
  while(n2%n_of_threads != 0)
    n_of_threads++;
  
  {
    //thread th[n_of_threads];
    for(int i = 0; i < n2; i+=n_of_threads*stack){
    
      vector<thread> th;
      
      if(output == 1)
	cout << "\rGenerating delta field " << number_of_grid+1 << ":       "  << data_type(400*i/n1)/10 + 40 << "%";
      else if(output == 2)
	cout << "\rGenerating kappa field " << number_of_grid+1 << ":       "  << data_type(200*i/n1)/10 + 40 << "%";
      else if(output == 3)
	cout << "\rGenerating delta field " << number_of_grid+1 << ":       "  << data_type(500*i/n1)/10 + 50 << "%";
      else if(output == 4)
	cout << "\rGenerating shear field " << number_of_grid+1 << ":       "  << data_type(300*i/n1)/10 + 70 << "%";
      cout.flush();

    
      for(int j = 0; j < n_of_threads; j++){
        th.push_back(thread(fft_aux_2,stack, i+stack*j, n1, f, e));
      }
    
      for(int j = 0; j < n_of_threads; j++){
        th[j].join();
      }
    
    }
  }
}


/*******************************************************************************************************************************************************
 * fft_2D_inverse
 * 
 * Description:
 * - inverse two dimensional Fourier transform
 * 
 * Arguments:
 * - f: pointer of the array to be transformed
 * - n1 & n2: edge sizes of the two dimensional array in (*f)
 * - thread_number: number of thread into which the computation
 *                  shall be split
 * - output: variable governing the output of the method
 * 
 * Comments:
 *   - The number of threads should be a power of 2 or at least n1 and n2 should be divisible by it.
 *     Otherwise an exception is drawn.
 * 
*******************************************************************************************************************************************************/


void fft_2D_inverse(valarray<valarray<complex<data_type> > > *f, int n1, int n2, int thread_number, int output, int number_of_grid){
  
  int n_of_threads;
  int stack = 8;
  
  complex<data_type> im(0,-constants::pi/data_type(n2));
  complex<data_type> e = exp(im);
  complex<data_type> normalization = complex<data_type>(data_type(1.0)/data_type(n1), 0.0);
	
  valarray<complex<data_type> > f1(0.0, n2);
  valarray<complex<data_type> > f2(0.0, n2);
  
  // Make sure, that n1 is divisible by number of threads.
  n_of_threads = thread_number;
  while(n1%n_of_threads != 0)
    n_of_threads++;
  
  valarray<valarray<complex<data_type> > > g(valarray<complex<data_type> >(0.0, n1), n_of_threads);
  {
    //thread th[n_of_threads];
    for(int i = 0; i < n1; i+=n_of_threads*stack){
      
      vector<thread> th;
      
      if(output == 1)
	cout << "\rGenerating kappa field " << number_of_grid+1 << ":       "  << data_type(200*i/n1)/10 + 60 << "%";
      else if(output == 2)
	cout << "\rGenerating log-normal field:    "  << data_type(250*i/n2)/10 + 50 << "%";
      cout.flush();
    
      for(int j = 0; j < n_of_threads; j++){
        th.push_back(thread(fft_aux_1_inverse,stack, i+stack*j, f, e, normalization));
      }
      
    
      for(int j = 0; j < n_of_threads; j++){
        th[j].join();
      }
    
    }
  }
  

  
  im = complex<data_type>(0,-constants::pi/data_type(n1));
  e = exp(im);
  normalization = complex<data_type>(data_type(1.0)/data_type(n2), 0.0);
  
  f1.resize(n1);
  f2.resize(n1);
  
  // Make sure, that n1 is divisible by number of threads.
  n_of_threads = thread_number;
  while(n2%n_of_threads != 0)
    n_of_threads++;
  
  g.resize(n_of_threads, valarray<complex<data_type> >(0.0, n1));
  {
    for(int i = 0; i < n2; i+=n_of_threads*stack){
    
      vector<thread> th;
      
      if(output == 1)
	cout << "\rGenerating kappa field " << number_of_grid+1 << ":       "  << data_type(200*i/n1)/10 + 80 << "%";
      else if(output == 2)
	cout << "\rGenerating log-normal field:    "  << data_type(250*i/n2)/10 + 50 << "%";
      cout.flush();
    
      for(int j = 0; j < n_of_threads; j++){
        th.push_back(thread(fft_aux_2_inverse,stack, i+stack*j, n1, f, e, normalization));
      }
    
      for(int j = 0; j < n_of_threads; j++){
        th[j].join();
      }
    
    }
  }
  
  
  
}