#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_cblas.h>


using namespace std;

/*
 * These function could as well be member functions of the class Grid, but 
 * the thread library doesn't accept member function in a thread.
 */

data_type apply_bias_in_log_normal_mode(double delta, double bias, double variance_of_Gaussian_field){
  
  if(delta <= -1.0)
    return -1.0;
  
  
  double x = log(1.0+bias*(exp(variance_of_Gaussian_field)-1.0))/variance_of_Gaussian_field;
  double n1 = log(1.0+delta)+variance_of_Gaussian_field/2.0;
  double n2 = n1*x-variance_of_Gaussian_field/2.0*x*x;
  
  return data_type(exp(n2)-1.0);
  
}


int draw_Poisson(data_type n_bar, data_type random_number){
 
  if(n_bar <= 0.0)
    return 0;
  int factoria = 1;
  int n = 0;
  data_type pow_n_bar = 1.0;
  data_type expo = exp(-n_bar);
  data_type u = random_number/expo;
  data_type upper_bound = 1.0;
  
  while(u > upper_bound){
    n++;
    factoria *= n;
    pow_n_bar *= n_bar;
    upper_bound += pow_n_bar/data_type(factoria);
  }
  
  return n;
  
}

void place_clustered_galaxies(vector<data_type> *positions, valarray<complex<data_type> > *grid, valarray<int> *number_grid, data_type dtheta, data_type n_bar, data_type bias, data_type bias_quadratic, double Gaussian_variance, int row, int offset, int log_normal, valarray<data_type> *random_numbers){
  
  int n = grid->size();
  int k;
  int index = 0;
  int n_of_rand = random_numbers->size();
  data_type mean = 0.0;
  data_type u;
  data_type v;
  data_type effective_bias;
  data_type delta;
  data_type delta_min;
  data_type dec_start, ra_start = (data_type(row-offset)-0.5)*dtheta;
  data_type ra, dec;
  data_type delta_N;
  
  // initializing the RNG. This is maybe still a bit hand-wavey and
  // should be improved..
  
  if(bias_quadratic == 0.0)
    delta_min = -data_type(1.0)/bias;
  else
    delta_min = (sqrt(bias*bias-bias_quadratic)-bias)/bias_quadratic;
  
  
  if(row >= offset && n-row > offset){
    for(int i = offset; i < n-offset; i++){
      delta = (*grid)[i].real();
      effective_bias = bias+bias_quadratic*delta;
      if(delta < delta_min)
	k = 0;
      else{
	
	if(log_normal == 1 && effective_bias != 1.0)
	  delta_N = apply_bias_in_log_normal_mode(double(delta), double(effective_bias), Gaussian_variance);
	else
	  delta_N = effective_bias*delta;
	k = draw_Poisson((1.0+delta_N)*n_bar, (*random_numbers)[index]);
        index = (index+1)%n_of_rand;
      }
      (*number_grid)[i] = k;
      //k = draw_Poisson((1.0+delta)*n_bar, (*random_numbers)[index]);
      dec_start = (data_type(i-offset)-0.5)*dtheta;
      for(int j = 0; j < k; j++){
	u = (*random_numbers)[index];
	index = (index+1)%n_of_rand;
	v = (*random_numbers)[index];
	index = (index+1)%n_of_rand;
	ra = ra_start + u*dtheta;
	dec = dec_start + v*dtheta;
	positions->push_back(ra);
	positions->push_back(dec);
      }
    }
  }
  
}


/*******************************************************************************************************************************************************
 * place_galaxies
 * 
 * Description:
 * - Interpolates the shear grid onto certain given galaxy positions.
 * - If "noise" is set, also intrinsic ellipticity noise is drawn for
 *   each galaxy.
 *
 * Arguments:
 * - (*positions): array storing the positions of the galaxies
 * - (*random_numbers): array storing the random numbers that will
 *                      be used to draw ellipticity noise
 * - (*thetas_de/ra): arrays storing the theta values that will be
 *                    used for interpolation
 * - (*grid): shear grid that will be interpolated onto the galaxy
 *            positions
 * - (*galaxy_data) array onto which the position, shear and noise
 *                  of each galaxy will be stored
 * - noise: switching noise on/off
 * - sigma: dispersion of the ellipticity noise
 * - output: governing the output of the method
 * - normalization: maximal random number
 *
 * 
*******************************************************************************************************************************************************/

void set_galaxy_data(vector<data_type> *positions, vector<data_type> *random_numbers,
		     valarray<data_type> *thetas_ra, valarray<data_type> *thetas_de, 
		     valarray<valarray<complex<data_type> > > * grid, vector<valarray<data_type> > * galaxy_data,
		     bool noise, data_type sigma, int output, data_type normalization){

  int n = (*positions).size()/2;
  int r = thetas_ra->size();
  int d = thetas_de->size();

  data_type ra;
  data_type de;
  data_type u1;
  data_type u2;
  data_type e1;
  data_type e2;
  
  complex<data_type> shear(0.0, 0.0);
  complex<data_type> intrinsic_noise(0.0, 0.0);
  
  for(int i = 0; i < n; i++){
    
    
    ra = (*positions)[2*i];
    de = (*positions)[2*i+1];
    
    shear = interpolate_complex_grid(ra, de, thetas_ra, thetas_de, grid, 3, 3);
    
    if(output == 0){
      cout << "\rPlacing galaxies:               "  << setprecision(1) << data_type(i)/data_type(n)*100.0 << "%";
      cout.flush();
    }
    
    if(noise){
      u1 = (*random_numbers)[2*i]/normalization;
      u2 = (*random_numbers)[2*i+1]/normalization;
      e1 = sqrt(-2*log(u1))*sigma;
      e2 = e1*cos(2*constants::pi*u2);
      e1 *= sin(2*constants::pi*u2);
      intrinsic_noise = complex<data_type>(e1, e2);
      shear = (shear + intrinsic_noise)/(complex<data_type>(1.0,0.0) + conj(shear)*intrinsic_noise);
    }
    
    (*galaxy_data)[i][0] = shear.real();
    (*galaxy_data)[i][1] = shear.imag();
    
  }
  
  if(output == 0){
    cout << "\rPlacing galaxies:               100%      \n";
  }
  
}


/*******************************************************************************************************************************************************
 * Still add documentation here!!!
 * 
*******************************************************************************************************************************************************/

void generate_random_numbers(int n, valarray<valarray<data_type> > *randoms, mt19937_64 *gen, data_type normalization, int field, int n_of_fields){
  for(int i = 0; i < n/2+1; i++){
    cout << "\rGenerating random numbers:      "  << data_type(2000*i/(n_of_fields*n) + 1000*field/(n_of_fields))/10 << "%";
    cout.flush();
    for(int j = 0; j < 2*n; j++){
      (*randoms)[i][j] = data_type((*gen)())/normalization;
    }
  }
}

void respect_symmetries(valarray<complex<data_type> >* grid1, valarray<complex<data_type> >* grid2){
          
  int n = grid1->size();  
  
    for(int j = 0; j < n; j++){
      if(j == 0)
        (*grid1)[j] = conj((*grid2)[0]);
      else
	(*grid1)[j] = conj((*grid2)[n-j]);
    }
  
}

void compute_k_modes(data_type dl, data_type l1, data_type l_min, data_type l_max,
		    valarray<data_type> ln_l_array, 
		    vector<vector<valarray<data_type> > > power_spectra,
		    vector< valarray<valarray<complex<data_type> > > >* grids,
		    int row,
		    vector<valarray<valarray<data_type> > > *random_numbers, 
		    vector<valarray<double> > *sigma){


  int n = (*random_numbers)[0][row].size()/2;
  int n_fields = power_spectra.size();
  
  data_type u1;
  data_type u2;
  data_type r;
  data_type im;
  data_type s;
  data_type l2;
  data_type l;
  data_type ln_l;
  data_type pow;
  data_type dV = dl*dl/(4*constants::pi_sq);
  
  // variables needed to transform modes into eigenspace of
  // mode covariance matrix, e.g. (Pd , Pcr; Pcr , Pk) for
  // two fields.
  vector<data_type> lambda(n_fields, 0.0);
  vector<complex<data_type> > v(n_fields, complex<data_type>(0.0));

  gsl_matrix * m = gsl_matrix_alloc(n_fields, n_fields);
  gsl_matrix * evec = gsl_matrix_alloc(n_fields, n_fields);
  gsl_vector * eval = gsl_vector_alloc(n_fields);
  
  for(int i = 0; i < n_fields; i++)
    (*sigma)[i][row] = 0.0;
  
  for(int i = 0; i < n; i++){
    
    
    if(i < n/2)
      l2 = data_type(i)*dl;
    else
      l2 = data_type(i)*dl - 2*l_max;
        
    l = sqrt(l1*l1+l2*l2);
    ln_l = log(l);
        
    if(l > l_max || l < l_min){
      for(int j = 0; j < n_fields; j++)
	(*grids)[j][row][i] = complex<data_type>(0.0, 0.0);
    }
    else{
      
      gsl_eigen_symmv_workspace *w = gsl_eigen_symmv_alloc(n_fields);
      for(int j = 0; j < n_fields; j++){
	for(int k = j; k < n_fields; k++){
	  pow = abs(interpolate_neville_aitken(ln_l, &ln_l_array, &power_spectra[j][k], 3));
	  gsl_matrix_set(m, j, k, pow);
	  gsl_matrix_set(m, k, j, pow);
	}
      }

      
      for(int j = 0; j < n_fields; j++)
	(*sigma)[j][row] += gsl_matrix_get(m, j, j);

      gsl_eigen_symmv(m, eval, evec, w);
      gsl_eigen_symmv_free(w);


      for(int j = 0; j < n_fields; j++){
       if(gsl_vector_get(eval, j) <= 0.0){
         //cerr << "Matrix of power spectra not positive definite at l = " << exp(ln_l) << ".\n";
         //exit(EXIT_FAILURE);
	  gsl_vector_set(eval, j, 0.0);
       }
	s = sqrt(gsl_vector_get(eval, j)/2.0)*dl/(2.0*constants::pi);
	u1 = (*random_numbers)[j][row][2*i];
	u2 = (*random_numbers)[j][row][2*i+1];
	r = sqrt(-2.0*log(u1))*s;
	im = r*cos(2.0*constants::pi*u2);
	r *= sin(2.0*constants::pi*u2);
	v[j] = complex<data_type>(r, im);
	
	for(int k = 0; k < n_fields; k++){
	  (*grids)[k][row][i] += v[j]*data_type(gsl_matrix_get(evec, k, j));
	}
      }

      // the total variance of the Gaussian fields has
      // to be computed for the log-normal transformation
      // that will follow later on.
      
    }
  }
  for(int j = 0; j < n_fields; j++)
    (*sigma)[j][row] *= dV;

  gsl_matrix_free(m);
  gsl_matrix_free(evec);
  gsl_vector_free(eval);
  
}




void transform_from_Gaussian_to_lognorm(valarray<complex<data_type> >* grid, double mu, data_type kappa_0){

  double kap;
  int n = grid->size();
  for(int i = 0; i < n; i++){
    kap = double((*grid)[i].real());
    (*grid)[i] = complex<data_type>(data_type(exp(kap+mu))-kappa_0, 0.0);
  }  
}


void apply_phase_factor(valarray<complex<data_type> >* grid, data_type dk, data_type l_max, int i){

  int n = grid->size();
  data_type k1, k2, k;
  data_type cos_phi, sin_phi;
  
  if(i < n/2)
    k1 = data_type(i)*dk;
  else
    k1 = data_type(i)*dk-2.0*l_max;
    
  for(int j = 0; j < n; j++){
    
    if(j < n/2)
        k2 = data_type(j)*dk;
      else
        k2 = data_type(j)*dk - 2*l_max;
  
      k = sqrt(k1*k1+k2*k2);
      
      if(k!=0){
        cos_phi = k1/k;
        sin_phi = k2/k;
        complex<data_type> phase(cos_phi*cos_phi-sin_phi*sin_phi,data_type(2.0)*sin_phi*cos_phi);
        (*grid)[j] *= phase;
      }
  }  
}


void reduce_shear_grid(valarray<complex<data_type> >* shear, valarray<complex<data_type> >* kappa){

  int n = shear->size();
  data_type kap;
    
  for(int i = 0; i < n; i++){
    (*shear)[i] /= data_type(1.0) - (*kappa)[i];
  }  
}
